// ESFrameObserverSwarm.java

import swarm.Globals;
import swarm.Selector;
import swarm.defobj.Zone;

import swarm.activity.Activity;
import swarm.activity.ActionGroup;
import swarm.activity.ActionGroupImpl;
import swarm.activity.Schedule;
import swarm.activity.ScheduleImpl;

import swarm.objectbase.Swarm;
import swarm.objectbase.VarProbe;
import swarm.objectbase.MessageProbe;
import swarm.objectbase.EmptyProbeMapImpl;

import swarm.simtoolsgui.GUISwarm;
import swarm.simtoolsgui.GUISwarmImpl;

import swarm.analysis.EZGraph;
import swarm.analysis.EZGraphImpl;

import swarm.gui.Colormap;
import swarm.gui.ColormapImpl;
import swarm.space.Value2dDisplay;
import swarm.space.Value2dDisplayImpl;
import swarm.gui.ZoomRaster;
import swarm.gui.ZoomRasterImpl;

import swarm.space.Object2dDisplay;
import swarm.space.Object2dDisplayImpl;

import swarm.collections.ListImpl;

import java.util.*;


/**
 * The Observer contains the Model and the graphic widgets
 * @author  Pietro Terna and several of his students
 */
public class ESFrameObserverSwarm extends GUISwarmImpl {

    /** update frequency */
    public int displayFrequency;

    /** the time at which we stop the simulation (if != 0) */
    public int timeToFinish;

    /** displaying all messages on the console */
    public boolean verboseChoice;

    /** two ActionGroup for sequence of GUI events */
    public ActionGroup displayActions1, displayActions2;

    /** the single Schedule instance */
    public Schedule displaySchedule;

    /** the Swarm we are observing */
    public ESFrameModelSwarm eSFrameModelSwarm;

    /** our graphics or EZGraph output to files*/
    public EZGraphImpl 
	totalTimeLengthGraph=null, numberOfUnitsGraph=null, productionGraph=null, 
	ratioGraph=null;
 
    // the raster where we see units and the related unitDisplay and zoomFactor */
    public ZoomRaster unitRaster;
    public Object2dDisplay unitDisplay;
    public int zoomFactor;

    // the unitActivityDisplay
    public Value2dDisplay unitActivityDisplay;

    /** the address of the probe, if we need to access it; remember casting
     *  (ESFrameObserverSwarm$1$ESFrameObserverProbeMap)
     */
    Object observerProbe;

    /** Constructor for class */
    public ESFrameObserverSwarm (Zone aZone) {

	super(aZone);

	// fill in the relevant parameters (only two, in this case).
	displayFrequency = 1;
	verboseChoice=false;
	zoomFactor=4;
	timeToFinish=0;

	// Build a customized probe map using a `local' subclass
	// (a special kind of Java `inner class') of the
	// EmptyProbeMapImpl class.
	// This building operation produces a separated class file, with
	// name  ESFrameObserverSwarm$1$ESFrameObserverProbeMap.class
        class ESFrameObserverProbeMap extends EmptyProbeMapImpl {
	    public VarProbe probeVariable (String name) {
		return
		    Globals.env.probeLibrary.getProbeForVariable$inClass
		    (name, ESFrameObserverSwarm.this.getClass ());
	    }
	    public MessageProbe probeMessage (String name) {
		return
		    Globals.env.probeLibrary.getProbeForMessage$inClass
		    (name, ESFrameObserverSwarm.this.getClass ());
	    }
	    public void addVar (String name) {
		addProbe (probeVariable (name));
	    }
	    public void addMessage (String name) {
		addProbe (probeMessage (name));
	    }

	    public ESFrameObserverProbeMap (Zone _aZone, Class aClass) {
		super (_aZone, aClass);
		addVar ("displayFrequency");
		addVar ("verboseChoice");
		addVar ("zoomFactor");
		addVar ("timeToFinish");
	    }
	}
	// Install our custom probeMap class directly into the
	// probeLibrary
	observerProbe = new ESFrameObserverProbeMap (aZone, getClass ());
	Globals.env.probeLibrary.setProbeMap$For
	    ((ESFrameObserverProbeMap) observerProbe, getClass ());
    }

    /** Create the objects used to display the model. */
    public Object buildObjects () {

	Colormap colormap;
	Selector sel;
	super.buildObjects ();
	
	System.out.println("NB, the left and rigth histogram bars are in unit of orders;");
	System.out.println("    the lines in the Production graph are in unit of steps.");

	// First, we create the model that we're actually observing. The
	// model is a subswarm of the observer.

	// creating the ESFrameModelSwarm via lispAppArchivers, which
	// refers to the jesframe.scm file where we can modify parameters
	// in a stable way without compiling again the code
	eSFrameModelSwarm =
	    (ESFrameModelSwarm)Globals.env.lispAppArchiver.getWithZone$key(
						      getZone(), "eSFrameModelSwarm");
	// instead of using the direct constructor way
	// eSFrameModelSwarm = new ESFrameModelSwarm (getZone ());

	// Now create probe objects on the model and ourselves. This gives a
	// simple user interface to let the user change parameters.

	Globals.env.createArchivedProbeDisplay (eSFrameModelSwarm,
						"eSFrameModelSwarm");
	Globals.env.createArchivedProbeDisplay (this, "eSFrameObserverSwarm");

	// Instruct the control panel to wait for a button event: we
	// halt here until someone hits a control panel button so the
	// user can get a chance to fill in parameters before the
	// simulation runs
	getControlPanel ().setStateStopped ();

	// to change the verboseChoice via probe you have to wright 'true'
	// or 'false'; not simply 't' or 'f'
	// setting the 'verbose' value in StartESFrame
	StartESFrame.verbose=verboseChoice;

	// First, let the model swarm build its objects.
	eSFrameModelSwarm.buildObjects ();

	// create a colormap, to be used with the raster to display units and their
	// activity
	colormap = new ColormapImpl(getZone());
	colormap.setColor$ToName((byte)0, "yellowgreen");
	colormap.setColor$ToName((byte)1, "black");
	colormap.setColor$ToName((byte)2, "yellow");
	colormap.setColor$ToName((byte)3, "blue");
	colormap.setColor$ToName((byte)4, "red");
	colormap.setColor$ToName((byte)5, "grey");

	// create a "raster widget", a 2-dimensional display
	// window.  We tell the raster what to do if it dies, give it
	// its colormap, set its zoom factor (its actual size on the
	// display screen), set its virtual dimensions to the size of
	// our world, and give it its title.
	unitRaster = new ZoomRasterImpl (getZone(), "unitRaster");
	sel = SwarmUtils.getSelector(this, "_unitRasterDeath_");
	unitRaster.enableDestroyNotification$notificationMethod 
		(this, sel);
	unitRaster.setColormap (colormap);
	unitRaster.setZoomFactor (zoomFactor);
	unitRaster.setWidth$Height((eSFrameModelSwarm.getUnitSpace()).
				   getSizeX(),
				   (eSFrameModelSwarm.getUnitSpace()).
				   getSizeY());
	unitRaster.setWindowTitle ("Unit Space");

	// This instructs the raster to digest all the information we
	// have just given it and to initialize itself for display.
	unitRaster.pack();

	// create an Object2dDisplay that will display the units
	// on the raster. (The Object2dDisplay relies on the objects to send
	// their own draw messages to the raster.)  Think of
	// the unitspace as overlaying the raster.)
	sel = SwarmUtils.getSelector("Unit", "drawSelfOn");
	unitDisplay = new Object2dDisplayImpl(getZone(), unitRaster, 
					     eSFrameModelSwarm.getUnitSpace(), sel);

	// the unit activity display
	unitActivityDisplay = new Value2dDisplayImpl(getZone(), unitRaster, 
					     colormap,
					     eSFrameModelSwarm.getUnitActivitySpace());


	// tell the Raster to send mouse clicks to the
	// unitDisplay, allowing the user to right-click on the display
	// to probe the individual unit. (The right button is button
	// number 3)
	unitRaster.setButton$Client$Message(3, unitDisplay, 
		   SwarmUtils.getSelector(unitDisplay, "makeProbeAtX$Y"));


	// Create the graph widget to display the total production time
	// for all the finished orders divided by their total recipe lengths
	/*
	totalTimeLengthGraph = new EZGraphImpl
	    (getZone (), "Ratio total time / total lengths", "Time",
	     "Ratio", "totalTimeLengthGraph");

	Globals.env.setWindowGeometryRecordName
	    (totalTimeLengthGraph, "totalTimeLengthGraph");

	// the data for the totalTimeLengthGraph
	totalTimeLengthGraph.createSequence$withFeedFrom$andSelector
	    ("ratio", eSFrameModelSwarm,
	     SwarmUtils.getSelector(eSFrameModelSwarm,"getTimeLengthRatio"));
	*/
       
	// create the graph of the number of units
	numberOfUnitsGraph = new EZGraphImpl
	    (getZone (), "Number of Units", "Time",
	     "num", "numberOfUnitsGraph");

	Globals.env.setWindowGeometryRecordName
	    (numberOfUnitsGraph, "numberOfUnitsGraph");

	// the data for the numberOfUnitsGraph
	numberOfUnitsGraph.createSequence$withFeedFrom$andSelector
	    ("numOfU", eSFrameModelSwarm,
	     SwarmUtils.getSelector(eSFrameModelSwarm,"getNumberOfUnits"));
	

	// create the production graph
	productionGraph = new EZGraphImpl
	    (getZone (), "Production", "Time",
	     "quantity", "productionGraph");

	Globals.env.setWindowGeometryRecordName
	    (productionGraph, "productionGraph");

	// the data for the productionGraph
	productionGraph.createSequence$withFeedFrom$andSelector
	    ("globalProduction", eSFrameModelSwarm,
	     SwarmUtils.getSelector(eSFrameModelSwarm,"getGlobalProduction"));
	productionGraph.createSequence$withFeedFrom$andSelector
	    ("producedFinalGoods", eSFrameModelSwarm,
	     SwarmUtils.getSelector(eSFrameModelSwarm,"getConcludedOrderSteps"));
	//	productionGraph.createSequence$withFeedFrom$andSelector
	//	    ("potentialProduction", eSFrameModelSwarm,
	//	     SwarmUtils.getSelector(eSFrameModelSwarm,"getPotentialProduction"));

	// create the graph of the ratios
	ratioGraph = new EZGraphImpl
	    (getZone (), "Ratios", "Time",
	     "Quotas", "ratioGraph");

	Globals.env.setWindowGeometryRecordName
	    (numberOfUnitsGraph, "ratioGraph");

	// the data for the ratioGraph
	ratioGraph.createSequence$withFeedFrom$andSelector
	    ("global/potential", this, SwarmUtils.getSelector(this,"getRatio1"));
	ratioGraph.createSequence$withFeedFrom$andSelector
	    ("final/potential", this, SwarmUtils.getSelector(this,"getRatio2"));
	ratioGraph.createSequence$withFeedFrom$andSelector
	    ("final/global", this, SwarmUtils.getSelector(this,"getRatio3"));

	return this;
    }
    

  /**
   * Create the actions necessary for the simulation. This is where
   * the schedule is built (but not run!)  Here we create a display
   * schedule - this is used to display the state of the world and
   * check for user input. 
   */

    public Object buildActions () {
	int i;

	super.buildActions();

	// First, let our model swarm build its own schedule.
	eSFrameModelSwarm.buildActions();

	// Create two ActionGroup for display
	displayActions1 = new ActionGroupImpl (getZone());
	displayActions2 = new ActionGroupImpl (getZone());

	// Add the methods to the ActionGroup

	// Schedule the update of the graphic widgets
	// EZGraph files
	/*
	displayActions1.createActionTo$message
	    (totalTimeLengthGraph,
	     SwarmUtils.getSelector(totalTimeLengthGraph, "step"));
	*/
	displayActions1.createActionTo$message
	    (numberOfUnitsGraph,
	     SwarmUtils.getSelector(numberOfUnitsGraph, "step"));
	displayActions1.createActionTo$message
	    (productionGraph,
	     SwarmUtils.getSelector(productionGraph, "step"));
	displayActions1.createActionTo$message
	    (ratioGraph,
	     SwarmUtils.getSelector(ratioGraph, "step"));

	// Schedule the update of the raster
	// unitActivityDisplay and unitDisplay
	displayActions1.createActionTo$message
	    (unitActivityDisplay,
	     SwarmUtils.getSelector(unitActivityDisplay, "display"));
	displayActions1.createActionTo$message
	    (unitDisplay,
	     SwarmUtils.getSelector(unitDisplay, "display"));
	displayActions1.createActionTo$message
	    (unitRaster,
	     SwarmUtils.getSelector(unitRaster, "drawSelf"));


	// Schedule the update of the probe displays
	displayActions2.createActionTo$message
	    (Globals.env.probeDisplayManager,
	     SwarmUtils.getSelector(Globals.env.probeDisplayManager,"update"));

	// Schedule an update for the whole user
	// interface code.  This is crucial: without this, no
	// graphics update and the control panel will be
	// dead. It's best to put it at the end of the display
	// schedule
	displayActions2.createActionTo$message(getActionCache(),
       		      SwarmUtils.getSelector(getActionCache(),"doTkEvents"));

	// check to finish
	displayActions2.createActionTo$message(this,
       		      SwarmUtils.getSelector(this,"finish"));

	// The display schedule. Note the repeat interval: display
	// is frequently the slowest part of a simulation, so redrawing less
	// frequently can be a help.
	displaySchedule = new ScheduleImpl (getZone (), displayFrequency);

	// insert ActionGroup instance on the repeating Schedule
	// instance

	for (i=0;i<displayFrequency;i++)
	    {
		if(i==displayFrequency-1)
		    displaySchedule.at$createAction (i, displayActions1);
		displaySchedule.at$createAction (i, displayActions2);
	    }
	
	return this;
    }

    /**
       activateIn: - activate the schedules so they're ready to run.
       The swarmContext argument has to do with what we were activated
       *in*.  Typically the ObserverSwarm is the top-level Swarm, so
       it's activated in "null". But other Swarms and Schedules and
       such will be activated inside of us.  */

    public Activity activateIn (Swarm swarmContext) {
	// First, activate ourselves (just pass along the context).
	super.activateIn (swarmContext);

	// Activate the model swarm in ourselves. The model swarm is a
	// subswarm of the observer swarm.
	eSFrameModelSwarm.activateIn (this);

	// Now activate our schedule in ourselves. This arranges for
	// the execution of the schedule we built.
	displaySchedule.activateIn (this);

	// Activate returns the swarm activity - the thing that's ready to run.
	return getActivity();
  }

    /** drop the Observer, but after having dropped the EZGraph */
    public void drop() {

	System.out.println("\nPress newly Quit to exit.");
	getControlPanel ().setStateStopped ();

	if(totalTimeLengthGraph!=null)
	    totalTimeLengthGraph.drop();
	totalTimeLengthGraph=null;

	if(numberOfUnitsGraph!=null)
	    numberOfUnitsGraph.drop();
	numberOfUnitsGraph=null;

	if(productionGraph!=null)
	    productionGraph.drop();
	productionGraph=null;

	unitRaster.drop ();

	super.drop();
    }

    /** to finish automatically, if timeToFinish!=0 */
    public void finish(){
	if(timeToFinish != 0 && timeToFinish<=Globals.env.getCurrentTime())
	    {
		System.out.println("jES virtually quitted at time " + 
				   Globals.env.getCurrentTime());
		getControlPanel ().setStateQuit ();;
	    }
    }

    // handling the raster
    public Object _unitRasterDeath_ (Object caller) {
	unitRaster.drop ();
	unitRaster = null;
	return this;
    }

    // evaluating ratio1 (global / potential production)
    public float getRatio1()
    {
	if(eSFrameModelSwarm.getPotentialProduction()==0)return 0;
	else return eSFrameModelSwarm.getGlobalProduction() /  
		 eSFrameModelSwarm.getPotentialProduction();
    }

    // evaluating ratio2 (global / potential production)
    public float getRatio2()
    {
	if (eSFrameModelSwarm.getPotentialProduction()==0)return 0;
	else return eSFrameModelSwarm.getConcludedOrderSteps() /  
		 eSFrameModelSwarm.getPotentialProduction();
    }

    // evaluating ratio2 (final / global production)
    public float getRatio3()
    {
	if(eSFrameModelSwarm.getGlobalProduction()==0)return 0;
	else return eSFrameModelSwarm.getConcludedOrderSteps() /  
		 eSFrameModelSwarm.getGlobalProduction();
    }

}
